import { generateComponentId } from '@previewjs/api';
import { svelte } from '@sveltejs/vite-plugin-svelte';
import fs from 'fs-extra';
import path from 'path';
import url from 'url';
import { maybeOptionalType, objectType } from '@previewjs/type-analyzer';
import ts from 'typescript';
import { ReaderListeners } from '@previewjs/vfs';
import { preprocess } from 'svelte/compiler';

function analyzeSvelteComponentFromSFC(typeAnalyzer, filePath) {
  const resolver = typeAnalyzer.analyze([filePath + ".ts"]);
  const sourceFile = resolver.sourceFile(filePath + ".ts");
  const propsTypeFields = {};
  let collected = {};
  for (const statement of sourceFile?.statements || []) {
    if (ts.isVariableStatement(statement) && statement.modifiers?.find(
      (m) => m.kind === ts.SyntaxKind.ExportKeyword
    ) && statement.declarationList.flags ^ ts.NodeFlags.Const) {
      for (const declaration of statement.declarationList.declarations) {
        if (!ts.isIdentifier(declaration.name)) {
          continue;
        }
        const { type: fieldType, collected: fieldCollected } = resolver.resolveType(
          resolver.checker.getTypeAtLocation(declaration.name)
        );
        propsTypeFields[declaration.name.text] = maybeOptionalType(
          fieldType,
          !!declaration.initializer
        );
        collected = { ...collected, ...fieldCollected };
      }
    }
  }
  return {
    propsType: objectType(propsTypeFields),
    types: collected
  };
}

function createSvelteTypeScriptReader(reader) {
  return new SvelteTypeScriptReader(reader);
}
class SvelteTypeScriptReader {
  constructor(reader) {
    this.reader = reader;
    this.listeners = new ReaderListeners();
    reader.listeners.add({
      onChange: (absoluteFilePath, info) => {
        this.listeners.notify(absoluteFilePath, info);
      }
    });
    this.observe = reader.observe?.bind(reader);
  }
  async read(filePath) {
    if (filePath.endsWith(".svelte.ts")) {
      const source = await this.reader.read(
        filePath.substr(0, filePath.length - 3)
      );
      if (source?.kind !== "file") {
        return null;
      }
      return {
        kind: "file",
        name: path.basename(filePath),
        realPath: async () => null,
        lastModifiedMillis: () => source.lastModifiedMillis(),
        read: async () => convertToTypeScript(await source.read()),
        size: () => source.size()
      };
    }
    const entry = await this.reader.read(filePath);
    if (entry?.kind === "directory") {
      return {
        ...entry,
        entries: async () => {
          const realEntries = await entry.entries();
          return Promise.all([
            ...realEntries.map((e) => this.read(path.join(filePath, e.name))),
            ...realEntries.filter((e) => e.name.endsWith(".svelte")).map((e) => this.read(path.join(filePath, e.name + ".ts")))
          ]).then((entries) => entries.filter(Boolean));
        }
      };
    }
    return entry;
  }
  readSync(filePath) {
    if (filePath.endsWith(".svelte.ts")) {
      const source = this.reader.readSync(
        filePath.substr(0, filePath.length - 3)
      );
      if (source?.kind !== "file") {
        return null;
      }
      return {
        kind: "file",
        name: path.basename(filePath),
        realPath: () => null,
        lastModifiedMillis: () => source.lastModifiedMillis(),
        read: () => convertToTypeScript(source.read()),
        size: () => source.size()
      };
    }
    const entry = this.reader.readSync(filePath);
    if (entry?.kind === "directory") {
      return {
        ...entry,
        entries: () => {
          const realEntries = entry.entries();
          return [
            ...realEntries.map(
              (e) => this.readSync(path.join(filePath, e.name))
            ),
            ...realEntries.filter((e) => e.name.endsWith(".svelte")).map((e) => this.readSync(path.join(filePath, e.name + ".ts")))
          ].filter(Boolean);
        }
      };
    }
    return entry;
  }
}
function convertToTypeScript(svelteTemplateSource) {
  let scriptContent = "";
  preprocess(svelteTemplateSource, [
    {
      script: ({ content }) => {
        scriptContent = content;
        return {
          code: ""
        };
      }
    }
  ]);
  return scriptContent;
}

const svelteFrameworkPlugin = {
  isCompatible: async (dependencies) => {
    const version = await dependencies["svelte"]?.readInstalledVersion();
    if (!version) {
      return false;
    }
    return parseInt(version) === 3;
  },
  async create({ rootDirPath }) {
    const __dirname = url.fileURLToPath(new URL(".", import.meta.url));
    const previewDirPath = path.resolve(__dirname, "..", "preview");
    return {
      pluginApiVersion: 3,
      name: "@previewjs/plugin-svelte",
      transformReader: (reader) => createSvelteTypeScriptReader(reader),
      defaultWrapperPath: "__previewjs__/Wrapper.svelte",
      previewDirPath,
      detectComponents: async (reader, typeAnalyzer, absoluteFilePaths) => {
        const components = [];
        for (const absoluteFilePath of absoluteFilePaths) {
          if (absoluteFilePath.endsWith(".svelte") && await fs.pathExists(absoluteFilePath)) {
            components.push({
              componentId: generateComponentId({
                filePath: path.relative(rootDirPath, absoluteFilePath),
                name: path.basename(
                  absoluteFilePath,
                  path.extname(absoluteFilePath)
                )
              }),
              offsets: [
                [0, (await fs.readFile(absoluteFilePath, "utf-8")).length]
              ],
              info: {
                kind: "component",
                exported: true,
                analyze: async () => analyzeSvelteComponentFromSFC(typeAnalyzer, absoluteFilePath)
              }
            });
          }
        }
        return components;
      },
      viteConfig: (configuredPlugins) => ({
        define: {
          __SVELTEKIT_DEV__: "false",
          __SVELTEKIT_APP_VERSION_POLL_INTERVAL__: "0"
        },
        publicDir: "static",
        resolve: {
          alias: {
            $app: "node_modules/@sveltejs/kit/src/runtime/app"
          }
        },
        plugins: [
          ...configuredPlugins.filter(
            (plugin) => plugin.name !== "vite-plugin-svelte-kit" && plugin.name !== "vite-plugin-sveltekit-build" && plugin.name !== "vite-plugin-sveltekit-middleware" && plugin.name !== "vite-plugin-sveltekit-setup" && plugin.name !== "vite-plugin-sveltekit-compile"
          ),
          configuredPlugins.find((plugin) => plugin.name.includes("svelte")) ? null : svelte(),
          {
            name: "previewjs:fake-sveltekit-client",
            transform(code, id) {
              if (id.includes("@sveltejs/kit/src/runtime/client/singletons.js")) {
                return code.replace(`export let client`, `export let client = {}`).replace(
                  `url: notifiable_store({})`,
                  `url: notifiable_store(document.location)`
                ).replace(
                  `page: notifiable_store({})`,
                  `page: notifiable_store({ url: document.location })`
                );
              }
              return null;
            }
          },
          {
            name: "previewjs:disable-svelte-hmr",
            async transform(code, id) {
              if (!id.endsWith(".svelte")) {
                return null;
              }
              return code.replace(/import\.meta\.hot/g, "false");
            }
          }
        ]
      })
    };
  }
};

export { svelteFrameworkPlugin as default };
